#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022 CASA project
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# Correspondence concerning CASA should be addressed as follows:
#        Internet email: casa-feedback@nrao.edu.

"""CASA Python Module

This is a standard python module that provides CASA tools and tasks
without regular CASA's bespoke CLI.
"""

import os
import re
import sys
import shutil
import platform
import subprocess
from subprocess import call as Proc
import glob

from setuptools import setup, Extension
from distutils.command.build import build
from setuptools.command.build_ext import build_ext

from urllib import request
from shutil import copyfileobj
from distutils.dir_util import copy_tree


casacpp_user_options = [
   ("with-casacpp=", None, "location where casacpp is installed")
]
mod_closure = False

class XmlCMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class XmlCMakeBuild(build):
    user_options = build.user_options + casacpp_user_options

    def initialize_options(self):
        super().initialize_options()
        self.with_casacpp = None

    def finalize_options(self):
        super().finalize_options()

    def run(self):
        super().run()

class XmlCMakeBuildExt(build_ext):
    user_options = build_ext.user_options + casacpp_user_options

    def initialize_options(self):
        super().initialize_options()
        self.with_casacpp = None
    
    def finalize_options(self):
        super().finalize_options()
        self.set_undefined_options("build", ("with_casacpp", "with_casacpp"))

    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: "
                               + ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def _xml_jar_fetch(self,xml_jar_path, xml_jar_url ):
        if not os.path.exists(os.path.dirname(xml_jar_path)):
            os.makedirs(os.path.dirname(xml_jar_path))
        if not os.path.exists( xml_jar_path ):
            with request.urlopen(xml_jar_url) as istream, open(xml_jar_path,'wb') as fd:
                copyfileobj( istream, fd )

    def _copy_tools_xml_pys(self):
        """Copy .py files generated by xml-casa to the build dir from where the
        wheel will be packaged."""
        # In /tmp/build-via-sdist-.../casatools-x.y.z, a directory like
        # build/lib.linux-x86_64-cpython-310/casatools
        # This uses distutils.command.build.build / self.build_lib
        # https://setuptools.pypa.io/en/latest/userguide/extension.html#setuptools.command.build.SubCommand.build_lib
        casatools_dir = os.path.join(self.build_lib, 'casatools')
        if not os.path.isdir(casatools_dir):
            raise RuntimeError(f'casatools build dir ({casatools_dir}) is not a directory!')
        print(f' * Found destination dir for SWIG generated tool modules / .py files: {casatools_dir}')
        print(f' * Copying SWIG generated files for every tool module into {casatools_dir}...')
        for fname in glob.glob('src/casatools/*.py'):
            print(f'Copying file {fname} -> {casatools_dir}')
            shutil.copy2(fname, casatools_dir)


    def _generate_code_from_xml(self):

        # XML parsing and binding generation
        # The code is in a jar file with a given version in a well-known URL
        xml_jar_file = 'xml-casa-assembly-1.86.jar'
        xml_jar_url = 'http://casa.nrao.edu/download/devel/xml-casa/java/%s' % xml_jar_file
        xml_jar_path = os.path.abspath(os.path.join( 'scripts', 'java', xml_jar_file))

        # Get the tarball
        self._xml_jar_fetch(xml_jar_path, xml_jar_url)

        # Set some paths for the generated code
        binding_generated_cpp = 'src/tools'
        binding_generated_swig = 'src/tools'
        binding_generated_python = 'src/casatools/'

        # The xml to be used to generate the code
        xml_sources = glob.glob('xml/*xml')

        print("Generating bindings from XML", xml_jar_path)

        # Generate the tool C++ code (headers and code)
        proc1 = subprocess.Popen(['java', '-jar', xml_jar_path, '-tgts', '-header',
                                  '-output-tool-subdir', 'output-h=%s' % binding_generated_cpp,
                                  'output-c=%s' % binding_generated_cpp ] + xml_sources,
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc1_out = proc1.communicate()
        if proc1.returncode != 0:
             print(proc1_out[0].decode(sys.getdefaultencoding(),"strict"))
             print(proc1_out[1].decode(sys.getdefaultencoding(),"strict"))
             sys.exit('casa xml C++ generation failed')

        # Generate the python code
        proc2 = subprocess.Popen(['java', '-jar', xml_jar_path, '-swig', '-tool',
                                  '-output-tool-subdir', 'output-i=%s' % binding_generated_swig,
                                  'output-tool=%s' % binding_generated_python ] + xml_sources,
                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc2_out = proc2.communicate()
        xml_gen_out = proc2_out[0].decode(sys.getdefaultencoding(),"strict")
        if proc2.returncode != 0:
             print(proc2_out[0].decode(sys.getdefaultencoding(),"strict"))
             print(proc2_out[1].decode(sys.getdefaultencoding(),"strict"))
             sys.exit('casa xml SWIG and python generation failed')
        print(xml_gen_out)

    def build_extension(self, ext):

        # Helper methods to determine if we are running in a virtual environment.
        # This is needed since '-DPython3_FIND_VIRTUALENV=FIRST' is broken and picks
        # Python versions outside of the virtual environment.
        def get_base_prefix_compat():
            """Get base/real prefix, or sys.prefix if there is none."""
            return getattr(sys, "base_prefix", None) or getattr(sys, "real_prefix", None) or sys.prefix

        def in_virtualenv():
            return get_base_prefix_compat() != sys.prefix 

        def isexe(f):
            return os.path.isfile(f) and os.access(f, os.X_OK)

        # Generate the SWIG *.i files, the tool python and the tool C++ code
        self._generate_code_from_xml()

        # Copy the python files generated by the XML tool
        self._copy_tools_xml_pys()
       
        # Required for auto-detection of auxiliary "native" libs
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        # Call cmake to compile the tools C++ code (as well as SWIG processing)
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPython3_EXECUTABLE=' + sys.executable]

        if in_virtualenv():
            cmake_args.append('-DPython3_FIND_VIRTUALENV=ONLY')

        if self.with_casacpp is not None:
            cmake_args.append('-DCASACPP_ROOT=' + self.with_casacpp)

        # Call cmake, first to create the Makefile and second to build everything
        # The CCACHE_BASEDIR CCACHE_NOHASHDIR env variables allow ccache
        # to work even if the absolute path of the compilation directory changes,
        # as is the cawe with the python "build" package, that creates a temp
        # dir under $TMPDIR/build-via-sdist-xxxxxx. See "man cmake" for details.
        sourcedir='src/tools'
        os.environ['CCACHE_BASEDIR'] = os.getcwd()
        os.environ['CCACHE_NOHASHDIR'] = 'true'
        if 'CMAKE_BUILD_PARALLEL_LEVEL' not in os.environ:
            cmake_par = os.sysconf('SC_NPROCESSORS_ONLN') -1
            os.environ['CMAKE_BUILD_PARALLEL_LEVEL'] = str(1 if cmake_par == 0 else cmake_par)
        subprocess.check_call(['cmake', sourcedir] + cmake_args)
        subprocess.check_call(['cmake', '--build', '.'])

        # Copy GCC libs on Macos. 
        # TODO: Make this work with non-"standard" gcc location
        if (sys.platform == 'darwin'):
            gcc_dir = "/opt/local/lib/libgcc/"
            casac_lib_dir = extdir + "/casatools/__casac__/lib"
            copy_tree(gcc_dir, casac_lib_dir)

    
        # Collect all the libraries in a private directory and set relative rpaths
        if mod_closure:
            if isexe("scripts/mod-closure"):
                print("generating module closure...")
                if Proc([ "scripts/mod-closure", "--preserve-relative-paths", extdir ]) != 0:
                    sys.exit("\tclosure generation failed...")
                if isexe("scripts/find-glibc-private"):
                    if Proc(["scripts/find-glibc-private", "--delete", extdir]) != 0:
                        sys.exit("\tGLIBC_PRIVATE cleanup failed...")

            #Convert /opt/local/lib to @rpath on macOS
            if (sys.platform == 'darwin'):
                casac_lib_dir = extdir + "/casatools/__casac__/lib"
                from subprocess import call
                status=call("for f in `ls`; do if [[ $(otool -l $f | grep \"path /opt/local/lib\" | grep -v  \"/opt/local/lib/.*gcc\") ]];then install_name_tool -rpath /opt/local/lib  @loader_path:@loader_path/lib:@rpath $f ;fi; done",cwd=casac_lib_dir,shell=True)
    

def generate_extensions():

    extensions = []
    cmdclass = {}

    cmdclass = {'build': XmlCMakeBuild, 'build_ext': XmlCMakeBuildExt}

    casatoolsswig_ext = XmlCMakeExtension('casatoolsswig')
    extensions.append(casatoolsswig_ext)

    return extensions, cmdclass

def compute_version():
    # If version.txt is found then use it. This would be the case if
    # building from source tarball, which is not longer a git repo.
    # Otherwise get the version using some logic on the git history and tags.
    if not os.path.exists("version.txt") :
        print("Getting version from git... ")
        with open("version.txt", "w") as version_stdout:
            proc = subprocess.Popen( [ "scripts/version" ], stdout=version_stdout, stderr=subprocess.PIPE )
            out,err = proc.communicate()
    with open("version.txt", "r") as version_stdout:
        out_version = version_stdout.readline()
        
    devbranchtag = out_version.split(" ")[0].strip() 
    releasetag = out_version.split(" ")[1].strip()
    dirty=""
    if (len(out_version.split(" ")) == 3): 
        print("Latest commit doesn't have a tag. Adding -dirty flag to version string.")
        dirty="+" + out_version.split(" ")[2].strip() # "+" denotes local version identifier as described in PEP440
    devbranchversion = ""
    devbranchrevision = ""
    if (devbranchtag != releasetag):
        devbranchrevision = devbranchtag.split("-")[-1]
        if (devbranchtag.startswith("CAS-")):
            devbranchversion=devbranchtag.split("-")[1] 
        else:
            devbranchversion=100
        devbranchrevision = devbranchtag.split("-")[-1]

    casatools_version = '%s%s' % (releasetag,dirty)
    if devbranchversion !="":
        casatools_version = '%sa%s.dev%s%s' % (releasetag,devbranchversion,devbranchrevision,dirty)
    print("casatools version: ", casatools_version)
    return casatools_version

def setup_package():

    META_DATA = dict(
        zip_safe=False,
        )

    # Set parameter 
    if '--mod-closure' in sys.argv:
        global mod_closure
        mod_closure = True
        mod_opt_pos = sys.argv.index('--mod-closure')
        sys.argv.pop(mod_opt_pos)
        if(mod_opt_pos < len(sys.argv)) :
            if sys.argv[mod_opt_pos] == 'yes' :
                sys.argv.pop(mod_opt_pos)


    if '--help' in sys.argv[1:] or \
      sys.argv[1] in ('--help-commands', 'egg_info', '--version'):
        pass
    else:
        # Generate extensions
        extensions, cmd_class = generate_extensions()

        META_DATA["version"] = compute_version()
        META_DATA["cmdclass"] = cmd_class
        META_DATA["ext_modules"] = extensions

    setup(**META_DATA)
    if os.path.exists("version.txt") and "sdist" in sys.argv:
        os.remove("version.txt")

if __name__ == '__main__':
    setup_package()
